/*BEGIN_LEGAL 
Intel Open Source License 

Copyright (c) 2002-2016 Intel Corporation. All rights reserved.
 
Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.  Redistributions
in binary form must reproduce the above copyright notice, this list of
conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.  Neither the name of
the Intel Corporation nor the names of its contributors may be used to
endorse or promote products derived from this software without
specific prior written permission.
 
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE INTEL OR
ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
END_LEGAL */
/*! @file
 *  This file contains an ISA-portable PIN tool for functional simulation of
 *  instruction+data TLB+cache hierarchies
 */

#include <iostream>
#include <fstream>
#include <stdio.h>

#include <stdlib.h>
#include <sys/time.h>


#include "pin.H"

typedef UINT32 CACHE_STATS; // type of cache hit/miss counters

// Benja: julio 17
FILE * trace; 
PIN_LOCK pinMutex;
PIN_LOCK esperaFin;
pid_t parent_pid;
int cnt=0;
static UINT64 icount = 0;
// Intervalo para imprimir estadisticas. -1 nunca (solo al final)
const UINT64 INSTR_PER_STATS = -1;
const char *prefijo;
const char *numOps;
UINT64 numOperMax;
UINT64 numOper=0;

int Assoc=1;
int Tam=16;

KNOB<string> KnobOutputFile(KNOB_MODE_WRITEONCE, "pintool",
        "o", "inscount.out", "specify output file name");

KNOB<string> KnobStopMemOps(KNOB_MODE_WRITEONCE, "pintool",
        "S", "60000000000", "specify number of instructions when detaching");

KNOB<string> KnobAssoc(KNOB_MODE_WRITEONCE, "pintool",
        "A", "1", "specify associativity of LLC");

KNOB<string> KnobTam(KNOB_MODE_WRITEONCE, "pintool",
        "T", "16", "specify size in MB of LLC");

#include "pin_cache.H"

namespace ITLB
{

    // instruction TLB: 4 kB pages, 32 entries, fully associative
    const UINT32 lineSize = 4*KILO;
    const UINT32 cacheSize = 32 * lineSize;
    const UINT32 associativity = 32;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);
    const UINT32 max_associativity = associativity;

    typedef CACHE_ROUND_ROBIN(max_sets, max_associativity, allocation) CACHE;
}
LOCALFUN ITLB::CACHE itlb("ITLB", ITLB::cacheSize, ITLB::lineSize, ITLB::associativity);

namespace DTLB
{
    // data TLB: 4 kB pages, 32 entries, fully associative
    const UINT32 lineSize = 4*KILO;
    const UINT32 cacheSize = 32 * lineSize;
    const UINT32 associativity = 32;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);
    const UINT32 max_associativity = associativity;

    typedef CACHE_ROUND_ROBIN(max_sets, max_associativity, allocation) CACHE;
}
LOCALVAR DTLB::CACHE dtlb("DTLB", DTLB::cacheSize, DTLB::lineSize, DTLB::associativity);

namespace IL1
{
    // 1st level instruction cache: 32 kB, 32 B lines, 32-way associative
    const UINT32 cacheSize = 32*KILO;
    const UINT32 lineSize = 32;
    const UINT32 associativity = 32;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_NO_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);
    const UINT32 max_associativity = associativity;

    typedef CACHE_ROUND_ROBIN(max_sets, max_associativity, allocation) CACHE;
}
LOCALVAR IL1::CACHE il1("L1 Instruction Cache", IL1::cacheSize, IL1::lineSize, IL1::associativity);

namespace DL1
{
    // 1st level data cache: 32 kB, 32 B lines, 32-way associative
    const UINT32 cacheSize = 32*KILO;
    const UINT32 lineSize = 32;
    const UINT32 associativity = 32;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_NO_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);
    const UINT32 max_associativity = associativity;

    typedef CACHE_ROUND_ROBIN(max_sets, max_associativity, allocation) CACHE;
}
LOCALVAR DL1::CACHE dl1("L1 Data Cache", DL1::cacheSize, DL1::lineSize, DL1::associativity);

namespace UL2
{
    // 2nd level unified cache: 2 MB, 64 B lines, direct mapped
    const UINT32 cacheSize = 2*MEGA;
    const UINT32 lineSize = 64;
    const UINT32 associativity = 1;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);

    typedef CACHE_DIRECT_MAPPED(max_sets, allocation) CACHE;
}
LOCALVAR UL2::CACHE ul2("L2 Unified Cache", UL2::cacheSize, UL2::lineSize, UL2::associativity);


namespace UL3
{
    // 3rd level unified cache: 16 MB, 64 B lines, direct mapped
    const UINT32 cacheSize = 16*MEGA;
    const UINT32 lineSize = 64;
    const UINT32 associativity = 1;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);

    typedef CACHE_DIRECT_MAPPED(max_sets, allocation) CACHE;
}
LOCALVAR UL3::CACHE ul3("L3 Unified Cache", UL3::cacheSize, UL3::lineSize, UL3::associativity);

namespace UL3AS8TAM16
{
    // 3rd level unified cache: 16 MB, 64 B lines, direct mapped
    const UINT32 cacheSize = 16*MEGA;
    const UINT32 lineSize = 64;
    const UINT32 associativity = 8;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);

    //typedef CACHE_DIRECT_MAPPED(max_sets, allocation) CACHE;
    typedef CACHE_ROUND_ROBIN(max_sets, associativity, allocation) CACHE;
}
LOCALVAR UL3AS8TAM16::CACHE ul3As8Tam16("L3 Unified Cache", UL3AS8TAM16::cacheSize, UL3AS8TAM16::lineSize, UL3AS8TAM16::associativity);

namespace UL3AS1TAM32
{
    // 3rd level unified cache: 16 MB, 64 B lines, direct mapped
    const UINT32 cacheSize = 32*MEGA;
    const UINT32 lineSize = 64;
    const UINT32 associativity = 1;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);

    typedef CACHE_DIRECT_MAPPED(max_sets, allocation) CACHE;
}
LOCALVAR UL3AS1TAM32::CACHE ul3As1Tam32("L3 Unified Cache", UL3AS1TAM32::cacheSize, UL3AS1TAM32::lineSize, UL3AS1TAM32::associativity);

namespace UL3AS8TAM32
{
    // 3rd level unified cache: 16 MB, 64 B lines, direct mapped
    const UINT32 cacheSize = 32*MEGA;
    const UINT32 lineSize = 64;
    const UINT32 associativity = 8;
    const CACHE_ALLOC::STORE_ALLOCATION allocation = CACHE_ALLOC::STORE_ALLOCATE;

    const UINT32 max_sets = cacheSize / (lineSize * associativity);

//    typedef CACHE_DIRECT_MAPPED(max_sets, allocation) CACHE;
    typedef CACHE_ROUND_ROBIN(max_sets, associativity, allocation) CACHE;
}
LOCALVAR UL3AS8TAM32::CACHE ul3As8Tam32("L3 Unified Cache", UL3AS8TAM32::cacheSize, UL3AS8TAM32::lineSize, UL3AS8TAM32::associativity);


LOCALFUN VOID Fini(int code, VOID * v)
{
    fprintf(stderr, "FINI NumberInstructions: %ld\n", icount);

    std::cerr << itlb;
    std::cerr << dtlb;
    std::cerr << il1;
    std::cerr << dl1;
    std::cerr << ul2;
      if ((Assoc==1) && (Tam==16))
        std::cerr << ul3;
      else if ((Assoc==8) && (Tam==16))
        std::cerr << ul3As8Tam16;
      else if ((Assoc==1) && (Tam==32))
        std::cerr << ul3As1Tam32;
      else 
        std::cerr << ul3As8Tam32;
    fflush(stderr);

    // Never ends. Protect the file trace
    PIN_GetLock(&pinMutex,0);
    cnt++;
    fprintf(stderr, "FINI cnt: %d\n", cnt);

    fclose(trace);
    PIN_ReleaseLock(&pinMutex);

}
LOCALFUN VOID imprime(UINT64 pp2)
{
UINT32 i;
UINT64 tmp;

  for (i=16; i>0;i--) {
    tmp = pp2 >> (i-1)*4;
    tmp = tmp & 0x000000000000000F;
    fprintf(trace, "%lx", tmp);
  }

}

LOCALFUN VOID ImprimeLinea(ADDRINT addrline, CACHE_BASE::ACCESS_TYPE accessType)
{
  UINT64 *pp, pp2;

// Lock is needed in multithread/multiprocess applications
        PIN_GetLock(&pinMutex,0);
// Using number of instructions instead of cycles
        if (accessType== CACHE_BASE::ACCESS_TYPE_LOAD)
          fprintf(trace, "%ld R %p ", icount, (void *)addrline);
        else if (accessType== CACHE_BASE::ACCESS_TYPE_REEMPLAZO)
          fprintf(trace, "%ld RE %p ", icount, (void *)addrline);
        else
          fprintf(trace, "%ld W %p ", icount, (void *)addrline);

        for (UINT32 i=0; i<8; i++) {
          pp = (UINT64 *)(addrline+i);
          // Si el acceso es de reemplazo no hay seguridad de que la direccion de memoria
          // accedida siga siendo valida por lo que no podemos acceder a ella. Si lo hiciera
          // se generaria un seg_fault de vez en cuando, cuando dicha zona hubiera sido
          // previamente liberada
          // Asigno valores 0f para no alterar la relacion entre 1s y 0s y asi evita 
          // alteraciones si hacemos un estudio vinculado a las celulas de memoria
          if (accessType== CACHE_BASE::ACCESS_TYPE_REEMPLAZO)
            pp2=0x0f0f0f0f0f0f0f0f;
          else
            pp2=*pp;
          imprime(pp2);
        } // for
        fprintf(trace," 0\n");
// Lock is needed in multithread/multiprocess application 
        fflush(trace);
        PIN_ReleaseLock(&pinMutex);

}

LOCALFUN VOID Ul2Access(ADDRINT addr, UINT32 size, CACHE_BASE::ACCESS_TYPE accessType)
{

    // second level unified cache
    const BOOL ul2Hit = ul2.Access(addr, size, accessType);

    // third level unified cache
    if ( ! ul2Hit)  {
      BOOL ul3Hit; 

      // Solo controlo reemplazos en L3 asi que si se han producido en L2 los ignoro
      Eviction=FALSE;

      if ((Assoc==1) && (Tam==16)) {
        ul3Hit = ul3.Access(addr, size, accessType);
      }
      else if ((Assoc==8) && (Tam==16)) {
        ul3Hit = ul3As8Tam16.Access(addr, size, accessType);
      }
      else if ((Assoc==1) && (Tam==32)) {
        ul3Hit = ul3As1Tam32.Access(addr, size, accessType);
      }
      else  {
        ul3Hit = ul3As8Tam32.Access(addr, size, accessType);
      }

      if (Eviction==TRUE) {
        Eviction=FALSE;
        ADDRINT addrlineR=(LineR<<6) & 0xFFFFFFFFFFC0;
//        printf ("-Ev LineR:%p ", (void *) LineR);
        ImprimeLinea(addrlineR, CACHE_BASE::ACCESS_TYPE_REEMPLAZO);
      }

// L3 misses are memory accesses
      if ( ! ul3Hit) {

        ADDRINT addrline = addr  & 0xFFFFFFFFFFC0;  // 6 bits menos significativos a 0
        ImprimeLinea(addrline, accessType);


      } // L3 miss

    }  // L2 miss
}



LOCALFUN VOID InsRef(ADDRINT addr)
{
    const UINT32 size = 1; // assuming access does not cross cache lines
    const CACHE_BASE::ACCESS_TYPE accessType = CACHE_BASE::ACCESS_TYPE_LOAD;

    icount ++;
    if (icount%INSTR_PER_STATS == 0) {
      fprintf(stderr, "NumberInstructions: %ld\n", icount);
      std::cerr << itlb;
      std::cerr << dtlb;
      std::cerr << il1;
      std::cerr << dl1;
      std::cerr << ul2;
      if ((Assoc==1) && (Tam==16))
        std::cerr << ul3;
      else if ((Assoc==8) && (Tam==16))
        std::cerr << ul3As8Tam16;
      else if ((Assoc==1) && (Tam==32))
        std::cerr << ul3As1Tam32;
      else 
        std::cerr << ul3As8Tam32;
    }


    // ITLB
    itlb.AccessSingleLine(addr, accessType);

    // first level I-cache
    const BOOL il1Hit = il1.AccessSingleLine(addr, accessType);

    // second level unified Cache
    if ( ! il1Hit) Ul2Access(addr, size, accessType);

    // Detach if max numops reached
    numOper++;
    if (numOper>=numOperMax)
      PIN_Detach();
}

LOCALFUN VOID MemRefMulti(ADDRINT addr, UINT32 size, CACHE_BASE::ACCESS_TYPE accessType)
{
    // DTLB
    dtlb.AccessSingleLine(addr, CACHE_BASE::ACCESS_TYPE_LOAD);

    // first level D-cache
    const BOOL dl1Hit = dl1.Access(addr, size, accessType);

    // second level unified Cache
    if ( ! dl1Hit) Ul2Access(addr, size, accessType);
}

LOCALFUN VOID MemRefSingle(ADDRINT addr, UINT32 size, CACHE_BASE::ACCESS_TYPE accessType)
{

    // DTLB
    dtlb.AccessSingleLine(addr, CACHE_BASE::ACCESS_TYPE_LOAD);

    // first level D-cache
    const BOOL dl1Hit = dl1.AccessSingleLine(addr, accessType);

    // second level unified Cache
    if ( ! dl1Hit) Ul2Access(addr, size, accessType);

}

LOCALFUN VOID Instruction(INS ins, VOID *v)
{
    // all instruction fetches access I-cache
    INS_InsertCall(
        ins, IPOINT_BEFORE, (AFUNPTR)InsRef,
        IARG_INST_PTR,
        IARG_END);

    if (INS_IsMemoryRead(ins) && INS_IsStandardMemop(ins))
    {
        const UINT32 size = INS_MemoryReadSize(ins);
        const AFUNPTR countFun = (size <= 4 ? (AFUNPTR) MemRefSingle : (AFUNPTR) MemRefMulti);

        // only predicated-on memory instructions access D-cache
        INS_InsertPredicatedCall(
            ins, IPOINT_BEFORE, countFun,
            IARG_MEMORYREAD_EA,
            IARG_MEMORYREAD_SIZE,
            IARG_UINT32, CACHE_BASE::ACCESS_TYPE_LOAD,
            IARG_END);
    }

    if (INS_IsMemoryWrite(ins) && INS_IsStandardMemop(ins))
    {
        const UINT32 size = INS_MemoryWriteSize(ins);
        const AFUNPTR countFun = (size <= 4 ? (AFUNPTR) MemRefSingle : (AFUNPTR) MemRefMulti);

        // only predicated-on memory instructions access D-cache
        INS_InsertPredicatedCall(
            ins, IPOINT_BEFORE, countFun,
            IARG_MEMORYWRITE_EA,
            IARG_MEMORYWRITE_SIZE,
            IARG_UINT32, CACHE_BASE::ACCESS_TYPE_STORE,
            IARG_END);
    }
}

VOID AfterForkParent(THREADID threadid, const CONTEXT* ctxt, VOID * arg)
{
      PIN_GetLock(&pinMutex, threadid+1);
        cerr << "TOOL: After fork parent." << endl;
        fprintf(stderr, "Ninst: %ld \n", icount); fflush(stderr);
      PIN_ReleaseLock(&pinMutex);
      parent_pid = PIN_GetPid();


//  Al crear un hijo el padre abandona pin ya que si crea más hijos los ciclos serán incorrectos
//     PIN_Detach();

}
VOID AfterForkChild(THREADID threadid, const CONTEXT* ctxt, VOID * arg)
{
      PIN_GetLock(&pinMutex, threadid+1);
        cerr << "TOOL: After fork child." << endl;
        fprintf(stderr, "Ninst: %ld \n", icount); fflush(stderr);

        // Cierro trace y abro uno nuevo
        fclose(trace);
        char nombre[10];
        char cadena[100];
        struct timeval timev;
        struct timezone timez;
        gettimeofday(&timev, &timez);
        srandom(timev.tv_usec);
        long int valor;
        int i;
        for (i=0;i<4;i++) {
          valor=random()*26/RAND_MAX;
          valor+=97;
          nombre[i]=valor;
        }
        nombre[4]='\0';

        sprintf(cadena,"%s_Ass%dTam%d_%s.nvt", prefijo, Assoc, Tam, nombre);
        printf("Fichero: %s\n", cadena);

trace=fopen(cadena, "w");

      PIN_ReleaseLock(&pinMutex);
      parent_pid = PIN_GetPid();



}


VOID BeforeFork(THREADID threadid, const CONTEXT* ctxt, VOID * arg)
{
      PIN_GetLock(&pinMutex, threadid+1);
        cerr << "TOOL: Before fork." << endl;
        fprintf(stderr, "Ninst: %ld \n", icount); fflush(stderr);
      PIN_ReleaseLock(&pinMutex);
      parent_pid = PIN_GetPid();


}

VOID ByeWorld(VOID *v)
{
      std::cerr << endl << "Detached at icount = " << icount << endl;
}


GLOBALFUN int main(int argc, char *argv[])
{
        char nombreA[10];
        char cadena[100];
        struct timeval timev;
        struct timezone timez;
        long int valor;
        int i;

    PIN_Init(argc, argv);

        numOps= KnobStopMemOps.Value().c_str();
        numOperMax=atol(numOps);
        prefijo= KnobOutputFile.Value().c_str();

        // Associativity and size of LLC
        Assoc=atoi(KnobAssoc.Value().c_str());
        Tam=atoi(KnobTam.Value().c_str());
        printf("Assoc: %d, Tam: %d-MB, Num.Max.Inst:%ld\n", Assoc, Tam, numOperMax);

        gettimeofday(&timev, &timez);
        srandom(timev.tv_usec);
        for (i=0;i<4;i++) {
          valor=random()*26/RAND_MAX;
          valor+=97;
          nombreA[i]=valor;
        }
        nombreA[4]='\0';
        sprintf(cadena,"%s_Ass%dTam%d_%s.nvt", prefijo, Assoc, Tam, nombreA);
        printf("Fichero: %s\n", cadena);



trace=fopen(cadena, "w");

    PIN_InitLock(&pinMutex);
    PIN_InitLock(&esperaFin);

    INS_AddInstrumentFunction(Instruction, 0);
    PIN_AddFiniFunction(Fini, 0);

    PIN_AddForkFunction(FPOINT_BEFORE, BeforeFork, 0);
    PIN_AddForkFunction(FPOINT_AFTER_IN_CHILD, AfterForkChild, 0);
    PIN_AddForkFunction(FPOINT_AFTER_IN_PARENT, AfterForkParent, 0);

    PIN_AddDetachFunction(ByeWorld, 0);

    // Never returns
    PIN_StartProgram();

    return 0; // make compiler happy
}
